# This file was auto-generated by Fern from our API Definition.

import typing

import pydantic
from ..core.pydantic_utilities import IS_PYDANTIC_V2, UniversalBaseModel
from .check_workflow_execution_status_error import CheckWorkflowExecutionStatusError
from .workflow_result_event_state import WorkflowResultEventState


class CheckWorkflowExecutionStatusResponse(UniversalBaseModel):
    """
    Response serializer for workflow execution status check.
    """

    status: WorkflowResultEventState
    outputs: typing.Optional[typing.Dict[str, typing.Optional[typing.Any]]] = None
    error: typing.Optional[CheckWorkflowExecutionStatusError] = None
    execution_id: str
    execution_detail_url: typing.Optional[str] = None

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow
