# This file was auto-generated by Fern from our API Definition.

import typing

from .node_execution_fulfilled_event import NodeExecutionFulfilledEvent
from .node_execution_rejected_event import NodeExecutionRejectedEvent

WorkflowSandboxExecuteNodeResponse = typing.Union[NodeExecutionFulfilledEvent, NodeExecutionRejectedEvent]
