# This file was auto-generated by Fern from our API Definition.

import typing

import pydantic
from ..core.pydantic_utilities import IS_PYDANTIC_V2, UniversalBaseModel
from .test_suite_run_prompt_sandbox_history_item_exec_config_data_request import (
    TestSuiteRunPromptSandboxHistoryItemExecConfigDataRequest,
)


class TestSuiteRunPromptSandboxHistoryItemExecConfigRequest(UniversalBaseModel):
    """
    Execution configuration for running a Test Suite against a Prompt Sandbox History Item
    """

    type: typing.Literal["PROMPT_SANDBOX_HISTORY_ITEM"] = "PROMPT_SANDBOX_HISTORY_ITEM"
    data: TestSuiteRunPromptSandboxHistoryItemExecConfigDataRequest
    test_case_ids: typing.Optional[typing.List[str]] = pydantic.Field(default=None)
    """
    Optionally specify a subset of test case ids to run. If not provided, all test cases within the test suite will be run by default.
    """

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow
