# This file was auto-generated by Fern from our API Definition.

import typing

import pydantic
from ..core.pydantic_utilities import IS_PYDANTIC_V2, UniversalBaseModel


class UpdateActiveWorkspaceResponse(UniversalBaseModel):
    update_active_workspace_id: typing.Optional[str] = pydantic.Field(default=None)
    """
    The id of the workspace that the user should update to, or null if no workspace change needed.
    """

    update_active_environment_id: typing.Optional[str] = pydantic.Field(default=None)
    """
    The id of the environment that the user should update to, or null if no environment change needed.
    """

    is_staff: typing.Optional[bool] = pydantic.Field(default=None)
    """
    Whether or not the user is a staff member of Vellum.
    """

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow
